#!/etc/opt/openit/perl

use strict;

use Fcntl qw( :DEFAULT :flock );
use File::Copy qw( copy );

print "Porting sge_bookmarks to new format.\n";

my $temp_dir = $ENV{TEMP_DIR};
if( ! defined $temp_dir ) {
    $temp_dir = `. /tmp/openit-cfg-shenv; echo \$TEMP_DIR`;
    chomp $temp_dir;
}

if( ! defined $temp_dir ) {
    die "ERROR: TEMP_DIR is not defined.\n";
}

my $sge_bookmarks = "$temp_dir/sge_bookmarks";
my $log_bookmarks = "$temp_dir/boookmarks.db";

if( ! -f $sge_bookmarks ) {
    print "No bookmarks of old format. Exiting.\n";
    exit( 0 );
}

# Check that everything is ready
my @check = ( "openit_clientd", "sgeacct_to_psm" );
foreach my $check( @check ) {
    my $output = `ps -eo args | grep $check | grep -v grep`;
    if( $output =~ /$check/ ) {
	die "ERROR: $check seems to be running.\n";
    }
}

my @r_bookmarks = get_bookmarks($sge_bookmarks);
my $bookmark_lines = 5;
my @new_bookmarks = ();

# Try to port every bookmark
foreach my $r_bookmark( @r_bookmarks ) {

    # Open file, if any
    my $file = $$r_bookmark{file};
    if( $file !~ m#^/# ) {
	die "ERROR: $file is not absolute path: Fix sge_bookmarks\n";
    }
    if( ! -f $file ) {
	print "No such file: Skipping $file\n";
	next;
    }
    unless( open( SGEFILE, "<$file" ) ) {
	print( "ERROR: Could not open $file. Skipping. $!\n" );
	next;
    }

    # Get needed bookmark info
    my $bm_line = $$r_bookmark{entry};
    chomp $bm_line;
    my $bm_time = $$r_bookmark{time};
    chomp $bm_time;

    my $line_ctr = 0; # Line counter
    my $bm_set = 0; # Have not set the bookmark yet
    my @new_bm_lines = ();

    $! = undef;
    $? = undef;
    while( <SGEFILE> ) {
	my $line = $_;
	chomp $line;

	# Save signature
	if( $line_ctr < $bookmark_lines ) {
	    push @new_bm_lines, $line;
	}

	# Skip non-matching lines
	my $time = ( split( ":", $line) )[ 10 ];
	if( $time < $bm_time ) {
	    $line_ctr++;
	    next;
	}
	if( $line ne $bm_line ) {
	    $line_ctr++;
	    next;
	}

	# Bookmark position found, make bookmark
	$line_ctr++;
	push @new_bookmarks,
	  $$r_bookmark{file} . ":$line_ctr:" . join( "@", @new_bm_lines );
	$bm_set = 1;
	last;
    }

    # Error checking
    if( $? ) {
	die "ERROR: Failed to read $file: $?\n";
    }
    close( SGEFILE );

    # Status info
    if( ! $bm_set ) {
	print "No bookmark for $file\n";
    } else {
	print "Located bookmark for $file\n";
    }
}

if( ! scalar @new_bookmarks ) {
    print "No bookmarks to update\n";
    if( -f $sge_bookmarks ) {
	print "Removing $sge_bookmarks\n";
	unlink $sge_bookmarks;
    }
    exit( 0 );
}

# Lock output file
print "Saving bookmarks with new format.\n";
open( LOCK_HANDLE, ">$log_bookmarks.lock" ) or
  die "Could not lock $log_bookmarks: $!\n";
if( ! flock( LOCK_HANDLE, LOCK_EX | LOCK_NB ) ) {
    my $err = $!;
    close( LOCK_HANDLE );
    unlink "$log_bookmarks.lock";
    die "Could not flock $log_bookmarks handle: $err\n";
}

if( -f $log_bookmarks && ! copy( "$log_bookmarks", "$log_bookmarks.tmp" ) ) {
    my $err = $!;
    unlink "$log_bookmarks.tmp";
    unlock();
    unlink "$log_bookmarks.lock";
    die "ERROR: Could not copy $log_bookmarks: $err\n";
}

# Write boookmark file
if( ! open( OUTPUT, ">>$log_bookmarks.tmp" ) ) {
    my $err = $!;
    unlock();
    unlink "$log_bookmarks.lock";
    die "ERROR: Failed to open $log_bookmarks.tmp: $err\n";
}

foreach my $new_bookmark( @new_bookmarks ) {
    if( ! print OUTPUT "$new_bookmark\n" ) {
	my $err = $!;
	# Rule out everything on error
	close( OUTPUT );
	unlink "$log_bookmarks.tmp";
	unlock();
	unlink "$log_bookmarks.lock";
	die "ERROR: Failed to write to $log_bookmarks: $err";
    }
}
close( OUTPUT );
if( ! rename( "$log_bookmarks.tmp", "$log_bookmarks" ) ) {
    my $err = $!;
    unlock();
    unlink "$log_bookmarks.lock";
    die "ERROR: Failed to rename $log_bookmarks.tmp." .
      " $log_bookmarks may be corrupt. $err\n";
}

# Unlock output file
unlock();
unlink "$log_bookmarks.lock";

print "Removing $sge_bookmarks\n";
unlink $sge_bookmarks;

sub unlock {
    if( ! flock( LOCK_HANDLE, LOCK_UN ) ) {
	print "WARNING: Unable to release lock on $log_bookmarks: $!\n";
    }
    close( LOCK_HANDLE );
}


#SUB##########################################################################
#
# @NAME:
#   get_bookmarks
# @PURPOSE:
#   Find bookmarks
# @ARGS:
#   * $bmfile  : object reference : FileObject for bookmark file (not opened)
# @RETURNS:
#   r_bookmarks: list of hashes   : Hashes with bookmark data, or empty
#
#/SUB

sub get_bookmarks {
    my $file = shift;
    my @r_bookmarks = ();

    my $r_bookmarks = read_bookmarks($file);

    foreach my $cell( keys %$r_bookmarks ) {
	push @r_bookmarks, $$r_bookmarks{$cell};
    }
    return @r_bookmarks;
}

#SUB###########################################################################
# @NAME:
#   load_bookmark
# @PURPOSE:
#   Read bookmarks from the specified bookmark file
#   Bookmarks consists of cell name, file path, timestamp and the
#   full record from the SGE accounting file in a colon-separated
#   tag="value" format.
# @ARGS:
#   * $bmfile : object reference : FileObject for bookmark file (not opened)
# @RETURNS:
#   bookmark  : hash reference   : Hash with bookmark data, undef if not found
#
#/SUB

sub read_bookmarks {
    my $file = shift;


    unless ( -e $file ) {
	return +{};
    }
    # Open file
    unless ( open( FILE, "<$file" ) ) {
	die "Unable to open bookmarks file\n";
    }

    my %bookmarks;

    # Read a line
    while( <FILE> ) {
	# Split the line:
	my $line = $_;
	my %bm;
	while ( $line =~ /(\w+)="(.+?)":?(.*)/ ) {
	    my $key = $1;
	    my $val = $2;
	    $line = $3;

	    $bm{$key} = $val;
	}

	$bookmarks{$bm{cell}}=\%bm;

    }
    close( FILE );


    return \%bookmarks;
}
